"""
Test module for the calibration module of the duet_tools package.
"""

from __future__ import annotations

import pytest
import numpy as np
from pathlib import Path
import math
import matplotlib.pyplot as plt
import geojson

from duet_tools.calibration import (
    DuetRun,
    Targets,
    FuelParameter,
    LandfireQuery,
    import_duet,
    assign_targets,
    set_fuel_parameter,
    set_density,
    set_moisture,
    set_height,
    calibrate,
    assign_targets_from_sb40,
    _maxmin_calibration,
)

from duet_tools.utils import read_dat_to_array
from duet_tools.landfire import query_landfire

TEST_DIR = Path(__file__).parent
TMP_DIR = TEST_DIR / "tmp"
TMP_DIR.mkdir(exist_ok=True)
DATA_DIR = TEST_DIR / "test-data"


class TestDuetRun:
    def test_import_duet(self):
        duet_run = import_duet(directory=DATA_DIR, nx=252, ny=252)
        # test that data types are correct
        assert isinstance(duet_run, DuetRun)
        assert isinstance(duet_run.density, np.ndarray)
        assert isinstance(duet_run.height, np.ndarray)
        assert duet_run.moisture is None
        # test array shapes
        assert duet_run.density.shape == (2, 252, 252)
        assert duet_run.height.shape == (2, 252, 252)
        # test that wrong dimensions raise error
        with pytest.raises(ValueError):
            duet_run = import_duet(directory=DATA_DIR, nx=252, ny=252, nz=3)

    def test_add_moisture(self):
        duet_run = import_duet(directory=DATA_DIR, nx=252, ny=252)
        grass_moist = np.full((252, 252), 0.4)
        litter_moist = np.full((252, 252), 0.1)
        array_to_add = np.array([grass_moist, litter_moist])
        duet_run.add_moisture_array(array_to_add)
        assert isinstance(duet_run.moisture, np.ndarray)
        assert duet_run.moisture.shape == (2, 252, 252)
        # test zero fuel moistue where fuels are present raises error
        with pytest.raises(ValueError):
            duet_run = import_duet(directory=DATA_DIR, nx=252, ny=252)
            duet_run.add_moisture_array(np.array([np.zeros((252, 252)), litter_moist]))
        # test wrong dimensions
        with pytest.raises(ValueError):
            duet_run = import_duet(directory=DATA_DIR, nx=252, ny=252)
            duet_run.add_moisture_array(grass_moist)

    def test_to_numpy(self):
        duet_run = import_duet(directory=DATA_DIR, nx=252, ny=252)
        grass_moist = np.full((252, 252), 0.4)
        litter_moist = np.full((252, 252), 0.1)
        array_to_add = np.array([grass_moist, litter_moist])
        duet_run.add_moisture_array(array_to_add)
        # test each fuel parameter and type
        grass_density = duet_run.to_numpy("grass", "density")
        litter_density = duet_run.to_numpy("litter", "density")
        grass_moisture = duet_run.to_numpy("grass", "moisture")
        litter_moisture = duet_run.to_numpy("litter", "moisture")
        grass_height = duet_run.to_numpy("grass", "height")
        litter_height = duet_run.to_numpy("litter", "height")
        assert np.array_equal(grass_density, duet_run.density[0, :, :])
        assert np.array_equal(litter_density, duet_run.density[1, :, :])
        assert np.array_equal(grass_moisture, duet_run.moisture[0, :, :])
        assert np.array_equal(litter_moisture, duet_run.moisture[1, :, :])
        assert np.array_equal(grass_height, duet_run.height[0, :, :])
        assert np.array_equal(litter_height, duet_run.height[1, :, :])
        # test integrated and separated
        separated_density = duet_run.to_numpy("separated", "density")
        assert np.array_equal(separated_density, duet_run.density)
        integrated_density = duet_run.to_numpy("integrated", "density")
        integrated_height = duet_run.to_numpy("integrated", "height")
        integrated_moisture = duet_run.to_numpy("integrated", "moisture")
        assert np.array_equal(integrated_density, np.sum(duet_run.density, axis=0))
        assert np.array_equal(integrated_height, np.max(duet_run.height, axis=0))
        weights = _maxmin_calibration(duet_run.density.copy(), max=1.0, min=0)
        weights[weights == 0] = 0.01
        weighted_average_moisture = np.average(
            duet_run.moisture, axis=0, weights=weights
        )
        assert np.array_equal(integrated_moisture, weighted_average_moisture)

    def test_to_quicfire(self):
        duet_run = import_duet(directory=DATA_DIR, nx=252, ny=252)
        duet_run.to_quicfire(TMP_DIR)
        treesrhof = read_dat_to_array(TMP_DIR, "treesrhof.dat", 252, 252, 1, order="C")
        treesrhof = treesrhof[0, :, :]
        treesfueldepth = read_dat_to_array(
            TMP_DIR, "treesfueldepth.dat", 252, 252, 1, order="C"
        )
        treesfueldepth = treesfueldepth[0, :, :]
        assert np.array_equal(treesrhof, duet_run._integrate("density"))
        assert np.array_equal(treesfueldepth, duet_run._integrate("height"))
        grass_moist = np.full((252, 252), 0.4)
        litter_moist = np.full((252, 252), 0.1)
        array_to_add = np.array([grass_moist, litter_moist])
        duet_run.add_moisture_array(array_to_add)
        duet_run.to_quicfire(TMP_DIR)
        treesmoist = read_dat_to_array(
            TMP_DIR, "treesmoist.dat", 252, 252, 1, order="C"
        )
        treesmoist = treesmoist[0, :, :]
        assert np.allclose(treesmoist, duet_run._integrate("moisture"))


class TestAssignTargets:
    def test_assign_targets(self):
        maxmin_targets = assign_targets(method="maxmin", max=1.0, min=0.2)
        meansd_targets = assign_targets(method="meansd", mean=0.6, sd=0.03)
        constant_target = assign_targets(method="constant", value=1.0)
        assert isinstance(maxmin_targets, Targets)
        assert isinstance(meansd_targets, Targets)
        assert isinstance(constant_target, Targets)
        assert maxmin_targets.targets == [1.0, 0.2]
        assert meansd_targets.targets == [0.6, 0.03]
        assert constant_target.targets == [1.0]
        # test other attributes
        assert maxmin_targets.args == ["max", "min"]
        assert maxmin_targets.method == "maxmin"
        assert maxmin_targets.calibration_function == _maxmin_calibration

    def test_target_validation(self):
        # test wrong method
        with pytest.raises(ValueError):
            assign_targets(method="minmax", min=0.2, max=1.0)
        # test wrong kwargs
        with pytest.raises(ValueError):
            assign_targets(method="maxmin", max=1.0)
        with pytest.raises(ValueError):
            assign_targets(method="constant", max=1.0, min=0.2)
        with pytest.raises(ValueError):
            assign_targets(method="meansd", max=1.0, min=0.02)
        # test incorrect inputs
        with pytest.raises(ValueError):
            assign_targets(method="maxmin", max=0.2, min=1.0)
        with pytest.warns(UserWarning):
            assign_targets(method="meansd", mean=0.03, sd=0.6)


class TestSetFuelParameter:
    def test_set_fuel_parameter(self):
        maxmin_targets = assign_targets(method="maxmin", max=1.0, min=0.2)
        meansd_targets = assign_targets(method="meansd", mean=0.6, sd=0.03)
        constant_target = assign_targets(method="constant", value=1.0)
        # test separated functions
        density_targets = set_density(grass=maxmin_targets, litter=meansd_targets)
        moisture_targets = set_moisture(grass=constant_target)
        height_targets = set_height(all=maxmin_targets)
        assert isinstance(density_targets, FuelParameter)
        assert isinstance(moisture_targets, FuelParameter)
        assert isinstance(height_targets, FuelParameter)
        assert density_targets.parameter == "density"
        assert moisture_targets.parameter == "moisture"
        # test generic function
        density_targets = set_fuel_parameter(
            parameter="density", grass=maxmin_targets, litter=meansd_targets
        )
        assert isinstance(density_targets, FuelParameter)
        assert density_targets.parameter == "density"
        assert density_targets.fuel_types == ["grass", "litter"]
        assert density_targets.targets == [maxmin_targets, meansd_targets]
        # test validation
        with pytest.raises(ValueError):
            density_targets = set_fuel_parameter(
                parameter="bulk density", litter=meansd_targets
            )
        with pytest.raises(ValueError):
            density_targets = set_fuel_parameter(
                parameter="density", both=meansd_targets
            )
        with pytest.raises(ValueError):
            density_targets = set_fuel_parameter(
                parameter="density", litter=meansd_targets, all=constant_target
            )


class TestCalibrate:
    def test_calibrate_maxmin(self):
        # try 1 fueltype and 1 parameter
        duet_run = import_duet(DATA_DIR, 252, 252)
        grass_density = assign_targets(method="maxmin", max=1.0, min=0.2)
        density_targets = set_fuel_parameter(parameter="density", grass=grass_density)
        calibrated_duet = calibrate(duet_run, fuel_parameter_targets=density_targets)
        assert isinstance(calibrated_duet, DuetRun)
        assert isinstance(calibrated_duet.density, np.ndarray)
        assert np.allclose(calibrated_duet.height, duet_run.height)
        assert np.allclose(calibrated_duet.density, duet_run.density) == False
        assert np.allclose(calibrated_duet.density[1, :, :], duet_run.density[1, :, :])
        assert np.max(calibrated_duet.density[0, :, :]) == np.float32(1.0)
        assert np.min(calibrated_duet.density[0, :, :]) == np.float32(0.2)
        # try density and height
        grass_height = assign_targets(method="maxmin", max=1.0, min=0.2)
        density_targets = set_fuel_parameter(parameter="density", grass=grass_density)
        height_targets = set_fuel_parameter(parameter="height", grass=grass_height)
        # can't calibrate height with maxmin because there's only one value!
        with pytest.raises(ValueError):
            calibrated_duet = calibrate(
                duet_run, fuel_parameter_targets=[density_targets, height_targets]
            )
        # now do moisture.. it will also raise an error since it doesn't exist in the og duet
        grass_moisture = assign_targets(method="maxmin", max=0.5, min=0.05)
        moisture_targets = set_fuel_parameter(
            parameter="moisture", grass=grass_moisture
        )
        with pytest.raises(ValueError):
            calibrated_duet = calibrate(
                duet_run, fuel_parameter_targets=moisture_targets
            )

        # try with two fueltypes
        litter_density = assign_targets(method="maxmin", max=0.1, min=0.01)
        density_targets = set_fuel_parameter(
            parameter="density", litter=litter_density, grass=grass_density
        )
        calibrated_duet = calibrate(duet_run, fuel_parameter_targets=density_targets)
        assert isinstance(calibrated_duet, DuetRun)
        assert isinstance(calibrated_duet.density, np.ndarray)
        assert np.allclose(calibrated_duet.height, duet_run.height)
        assert np.allclose(calibrated_duet.density, duet_run.density) == False
        assert np.max(calibrated_duet.density[0, :, :]) == np.float32(1.0)
        assert np.max(calibrated_duet.density[1, :, :]) == np.float32(0.1)

    def test_calibrate_meansd(self):
        # try 1 fueltype and 1 parameter
        duet_run = import_duet(DATA_DIR, 252, 252)
        grass_density = assign_targets(method="meansd", mean=0.6, sd=0.3)
        density_targets = set_fuel_parameter(parameter="density", grass=grass_density)
        calibrated_duet = calibrate(duet_run, fuel_parameter_targets=density_targets)
        assert isinstance(calibrated_duet, DuetRun)
        assert isinstance(calibrated_duet.density, np.ndarray)
        assert np.allclose(calibrated_duet.height, duet_run.height)
        assert np.allclose(calibrated_duet.density, duet_run.density) == False
        assert np.allclose(calibrated_duet.density[1, :, :], duet_run.density[1, :, :])
        assert math.isclose(
            np.mean(calibrated_duet.density[0, :, :]), np.float32(0.6), abs_tol=10**-6
        )
        assert math.isclose(
            np.std(calibrated_duet.density[0, :, :]), np.float32(0.3), abs_tol=10**-6
        )
        # try density and height
        grass_height = assign_targets(method="meansd", mean=0.5, sd=0.05)
        density_targets = set_fuel_parameter(parameter="density", grass=grass_density)
        height_targets = set_fuel_parameter(parameter="height", grass=grass_height)
        # shouldn't calibrate height with meansd because there's only one value
        with pytest.raises(ValueError):
            calibrated_duet = calibrate(
                duet_run, fuel_parameter_targets=[density_targets, height_targets]
            )

        # try with two fueltypes
        litter_density = assign_targets(method="meansd", mean=0.05, sd=0.005)
        density_targets = set_fuel_parameter(
            parameter="density", litter=litter_density, grass=grass_density
        )
        calibrated_duet = calibrate(duet_run, fuel_parameter_targets=density_targets)
        assert isinstance(calibrated_duet, DuetRun)
        assert isinstance(calibrated_duet.density, np.ndarray)
        assert np.allclose(calibrated_duet.height, duet_run.height)
        assert np.allclose(calibrated_duet.density, duet_run.density) == False

    def test_constant_calibration(self):
        duet_run = import_duet(DATA_DIR, 252, 252)
        grass_height = assign_targets(method="constant", value=0.5)
        litter_height = assign_targets(method="constant", value=0.05)
        height_targets = set_fuel_parameter(
            parameter="height", grass=grass_height, litter=litter_height
        )
        calibrated_duet = calibrate(duet_run, fuel_parameter_targets=height_targets)
        assert isinstance(calibrated_duet.height, np.ndarray)
        assert math.isclose(
            np.mean(
                calibrated_duet.height[0, :, :][calibrated_duet.height[0, :, :] > 0]
            ),
            np.float32(0.5),
            abs_tol=10**-6,
        )
        assert math.isclose(
            np.mean(
                calibrated_duet.height[1, :, :][calibrated_duet.height[1, :, :] > 0]
            ),
            np.float32(0.05),
            abs_tol=10**-6,
        )

    def test_fueltype_all(self):
        duet_run = import_duet(DATA_DIR, 252, 252)
        density = assign_targets(method="maxmin", max=2.0, min=0.5)
        grass_height = assign_targets(method="constant", value=0.75)
        litter_height = assign_targets(method="constant", value=0.15)
        density_targets = set_fuel_parameter(parameter="density", all=density)
        height_targets = set_fuel_parameter(
            parameter="height", grass=grass_height, litter=litter_height
        )
        calibrated_duet = calibrate(
            duet_run,
            fuel_parameter_targets=[density_targets, height_targets],
        )
        assert isinstance(calibrated_duet, DuetRun)
        # first assert that the height values are what we set them to be
        assert math.isclose(
            np.mean(
                calibrated_duet.height[0, :, :][calibrated_duet.height[0, :, :] > 0]
            ),
            np.float32(0.75),
            abs_tol=10**-6,
        )
        assert math.isclose(
            np.mean(
                calibrated_duet.height[1, :, :][calibrated_duet.height[1, :, :] > 0]
            ),
            np.float32(0.15),
            abs_tol=10**-6,
        )
        # then assert that a) the density array has two layers, and b) the overall max and min matches
        assert calibrated_duet.density.shape[0] == 2
        assert np.max(np.sum(calibrated_duet.density, axis=0)) == np.float32(2.0)
        assert np.min(np.sum(calibrated_duet.density, axis=0)) == np.float32(0.5)


class TestLandfireTargets:
    @classmethod
    def get_geojson(self):
        with open(DATA_DIR / "anderson_butte.geojson") as fid:
            sample_geojson = geojson.load(fid)
        return geojson.Polygon(sample_geojson["features"][0]["geometry"]["coordinates"])

    def test_query_landfire(self):
        sample_aoi = self.get_geojson()
        query = query_landfire(
            area_of_interest=sample_aoi, directory=TMP_DIR, input_epsg=4326
        )
        assert isinstance(query, LandfireQuery)
        assert isinstance(query.fuel_types, np.ndarray)
        assert isinstance(query.density, np.ndarray)
        assert isinstance(query.moisture, np.ndarray)
        assert isinstance(query.height, np.ndarray)

    def test_assign_targets_from_sb40(self):
        sample_aoi = self.get_geojson()
        query = query_landfire(
            area_of_interest=sample_aoi, directory=TMP_DIR, input_epsg=4326
        )
        # test just grass density
        grass_density = assign_targets_from_sb40(query, "grass", "density")
        assert isinstance(grass_density, Targets)
        assert grass_density.method == "maxmin"
        assert grass_density.args == ["max", "min"]
        assert len(grass_density.targets) == 2
        # test fuel and parameter with only one value
        with pytest.warns(UserWarning):
            grass_moisture = assign_targets_from_sb40(query, "grass", "moisture")
        grass_moisture = assign_targets_from_sb40(
            query, "grass", "moisture", method="constant"
        )
        # get the rest of the fuels and params with maxmin
        grass_height = assign_targets_from_sb40(query, "grass", "height")
        litter_density = assign_targets_from_sb40(query, "litter", "density")
        litter_moisture = assign_targets_from_sb40(query, "litter", "density")
        litter_height = assign_targets_from_sb40(query, "litter", "density")
        all_density = assign_targets_from_sb40(query, "all", "density")
        all_moisture = assign_targets_from_sb40(query, "all", "moisture")
        all_height = assign_targets_from_sb40(query, "all", "height")
        # test a couple with meansd
        grass_density_meansd = assign_targets_from_sb40(
            query, "grass", "density", method="meansd"
        )
        all_height = assign_targets_from_sb40(query, "all", "height", "meansd")
        # test wrong inputs
        with pytest.raises(ValueError):
            grass_height = assign_targets_from_sb40(query, "both", "height")
        with pytest.raises(ValueError):
            grass_height = assign_targets_from_sb40(query, "all", "moist")
        with pytest.raises(ValueError):
            grass_height = assign_targets_from_sb40(
                query, "all", "height", method="minmax"
            )


def plot_array(x, title):
    plt.figure(2)
    plt.set_cmap("viridis")
    plt.imshow(x, origin="lower")
    plt.colorbar()
    plt.title(title, fontsize=18)
    plt.show()



from duet_tools.calibration import (
    DuetRun,
    Targets,
    FuelParameter,
    import_duet,
    assign_targets,
    set_density,
    set_moisture,
    set_height,
    set_fuel_parameter,
    calibrate,
    assign_targets_from_sb40,
)

from duet_tools.landfire import (
    LandfireQuery,
    query_landfire,
)

from duet_tools.utils import (
    write_array_to_dat,
    read_dat_to_array,
)


__all__ = [
    "DuetRun",
    "Targets",
    "FuelParameter",
    "import_duet",
    "assign_targets",
    "set_density",
    "set_moisture",
    "set_height",
    "set_fuel_parameter",
    "calibrate",
    "assign_targets_from_sb40",
    "LandfireQuery",
    "query_landfire",
    "write_array_to_dat",
    "read_dat_to_array",
]



"""
Functions for interfacing with the LandFire API and processing the outputs
"""

# Core imports
from pathlib import Path
import importlib.resources
import re
import zipfile

# External imports
import numpy as np
import pandas as pd
import geojson
import shapely
from pyproj import Transformer
import landfire
from landfire.geospatial import get_bbox_from_polygon
import rasterio as rio

try:  # Python 3.9+
    DATA_PATH = importlib.resources.files("duet_tools").joinpath("data")
except AttributeError:  # Python 3.6-3.8
    from pkg_resources import resource_filename

    DATA_PATH = resource_filename("duet_tools", "data")


class LandfireQuery:
    """
    Class containing the information from a LandFire query, to be passed to assign_targets()
    """

    def __init__(
        self,
        fuel_types: np.ndarray,
        density: np.ndarray,
        moisture: np.ndarray,
        height: np.ndarray,
    ):
        self.fuel_types = fuel_types
        self.density = density
        self.moisture = moisture
        self.height = height
        self._validate_arrays_shape()

    def _get_fueltype_indices(self, arr: np.ndarray, ft: int):
        ft_dict = {1: "grass", -1: "litter"}
        if ft not in arr.flatten():
            raise ValueError(f"Landfire query does not contain any {ft_dict[ft]} fuel.")
        return np.where(arr == ft)

    def _validate_arrays_shape(self):
        attributes = [
            getattr(self, attr)
            for attr in dir(self)
            if not callable(getattr(self, attr)) and not attr.startswith("__")
        ]
        first_shape = None
        for arr in attributes:
            if isinstance(arr, np.ndarray):
                if first_shape is None:
                    first_shape = arr.shape
                elif arr.shape != first_shape:
                    raise ValueError(
                        "All arrays in LandfireQuery must have the same shape."
                    )

    def _validate_get_targets(self, fuel_type, parameter, method):
        fueltypes_allowed = ["grass", "litter", "all"]
        parameters_alowed = ["density", "moisture", "height"]
        methods_allowed = ["maxmin", "meansd", "constant"]

        if fuel_type not in fueltypes_allowed:
            raise ValueError(
                f"Fuel type {fuel_type} not supported. Must be one of {fueltypes_allowed}."
            )
        if parameter not in parameters_alowed:
            raise ValueError(
                f"Parameter {parameter} not supported. Must be one of {parameters_alowed}."
            )
        if method not in methods_allowed:
            raise ValueError(
                f"Method {method} not supported. Must be one of {methods_allowed}."
            )


def query_landfire(
    area_of_interest: geojson.Polygon | shapely.Polygon,
    directory: str | Path,
    input_epsg: int,
    delete_files: bool = True,
) -> LandfireQuery:
    """
    Creates and submits a LANDFIRE query for a specified area of interest.

    Parameters
    ----------
    area_of_interest : geojson.Polygon | shapely.Polygon
        Area in which to query LANDFIRE data. For best results, dimensions in meters should
        match (nx*dx, ny*dy) of DUET domain.
    directory : Path | str
        Directory where files associated with the LANDFIRE query will be saved.
    input_epsg : int
        EPSG number for CRS of area_of_interest polyong
    delete_files : bool = True
        Whether to delete intermediate files created in the process of querying LANDFIRE data. Defaults to True

    Returns
    -------
    LandfireQuery
    """
    if isinstance(directory, str):
        directory = Path(directory)

    if isinstance(area_of_interest, geojson.Polygon):
        area_of_interest = shapely.Polygon(area_of_interest["coordinates"][0])

    if input_epsg != 4236:
        area_of_interest = _reproject_polygon(
            area_of_interest, input_epsg, target_epsg=4326
        )

    _query_landfire(poly=area_of_interest, directory=directory)
    landfire_arr = _landfire_to_array(directory)

    # Import SB40 FBFM parameters table
    sb40_params_path = DATA_PATH / "sb40_parameters.csv"
    sb40_params = pd.read_csv(sb40_params_path)

    # Generate dict of fastfuels bulk density values and apply to Landfire query
    sb40_dict = _get_sb40_fuel_params(sb40_params)
    sb40_arr = _get_sb40_arrays(landfire_arr, sb40_dict)

    if delete_files:
        _delete_intermediate_files(directory)

    return LandfireQuery(
        fuel_types=sb40_arr[0, :, :],
        density=sb40_arr[1, :, :],
        moisture=sb40_arr[2, :, :],
        height=sb40_arr[3, :, :],
    )


def _query_landfire(
    poly: shapely.Polygon,
    directory: Path,
) -> None:
    """
    Download a grid of SB40 fuel models from Landfire for the unit and convert to a numpy array
    """

    bbox = get_bbox_from_polygon(aoi_polygon=poly, crs=4326)

    # Download Landfire data to output directory
    lf = landfire.Landfire(bbox, output_crs="5070")
    lf.request_data(
        layers=["200F40_19"], output_path=Path(directory, "landfire_sb40.zip")
    )

    # Exctract tif from compressed download folder and rename
    with zipfile.ZipFile(Path(directory, "landfire_sb40.zip")) as zf:
        extension = ".tif"
        rename = "landfire_sb40.tif"
        info = zf.infolist()
        for file in info:
            if file.filename.endswith(extension):
                file.filename = rename
                zf.extract(file, directory)


def _landfire_to_array(
    directory: Path,
) -> np.ndarray:
    # Upsample landfire raster to the quicfire resolution
    with rio.open(Path(directory, "landfire_sb40.tif")) as sb:
        arr = sb.read(1)

    return arr


def _get_sb40_fuel_params(params: pd.DataFrame) -> dict:
    """
    Builds a dictionary of SB40 fuel parameter values and converts them to
    the official FastFuels units

    Returns:
        dict: SB40 parameters for each fuel model
    """

    # Convert tons/ac-ft to kg/m^3
    params["1_hr_kg_per_m3"] = params["1_hr_t_per_ac"] * 0.22417
    params["10_hr_kg_per_m3"] = params["10_hr_t_per_ac"] * 0.22417
    params["100_hr_kg_per_m3"] = params["100_hr_t_per_ac"] * 0.22417
    params["live_herb_kg_per_m3"] = params["live_herb_t_per_ac"] * 0.22417
    params["live_woody_kg_per_m3"] = params["live_woody_t_per_ac"] * 0.22417

    # Convert inverse feet to meters
    params["dead_1_hr_sav_ratio_1_per_m"] = (
        params["dead_1_hr_sav_ratio_1_per_ft"] * 3.2808
    )
    params["live_herb_sav_ratio_1_per_m"] = (
        params["live_herb_sav_ratio_1_per_ft"] * 3.2808
    )
    params["live_wood_sav_ratio_1_per_m"] = (
        params["live_wood_sav_ratio_1_per_ft"] * 3.2808
    )

    # Convert percent to ratio
    params["dead_fuel_extinction_moisture"] /= 100

    # Convert feet to meters
    params["fuel_bed_depth_m"] = params["fuel_bed_depth_ft"] * 0.3048

    # Compute wet loading
    params["wet_load"] = params["1_hr_kg_per_m3"] + params["live_herb_kg_per_m3"]

    # Compute a live herb curing factor alpha as a function of wet loading.
    # This is kind of a B.S. approach raised by Rod on a phone call with
    # Anthony on 02/28/2023. I don't like this at all, but it is a temporary
    # Fix for the BP3D team to run some simulations.
    # low_load_fuel_models = [
    params["alpha"] = [0.5 if rho > 1 else 1.0 for rho in params["wet_load"]]

    # Compute dry loading
    params["dry_herb_load"] = params["live_herb_kg_per_m3"] * params["alpha"]
    params["dry_load"] = params["1_hr_kg_per_m3"] + params["dry_herb_load"]

    # Compute SAV
    params["sav_1hr_ratio"] = params["1_hr_kg_per_m3"] / params["dry_load"]
    params["sav_1hr"] = params["sav_1hr_ratio"] * params["dead_1_hr_sav_ratio_1_per_m"]
    params["sav_herb_ratio"] = params["dry_herb_load"] / params["dry_load"]
    params["sav_herb"] = (
        params["sav_herb_ratio"] * params["live_herb_sav_ratio_1_per_m"]
    )
    params["sav"] = params["sav_1hr"] + params["sav_herb"]

    # Convert nan to 0
    params.fillna(0, inplace=True)

    # Create dictionary for assigning fuel types for DUET calibration
    duet_dict = {
        "NB": 0,  # 0 = NEUTRAL, i.e. not predominantly grass or litter
        "GR": 1,  # 1 = GRASS predominantly
        "GS": 1,
        "SH": 1,  # I am considering shrubs as grass
        "TU": 0,
        "TL": -1,  # -1 = LITTER predominantly
        "SB": 0,
    }

    # Add column to df with DUET designations
    pattern = r"[0-9]"  # take out numbers from fbfm_type strings
    params["fbfm_cat"] = params["fbfm_code"].apply(lambda x: re.sub(pattern, "", x))
    params["duet_fuel_type"] = params["fbfm_cat"].apply(lambda x: duet_dict.get(x))

    # Build the dictionary with fuel parameters for the Scott and Burgan 40
    # fire behavior fuel models. Dict format: key ->
    # [name, loading (tons/ac), SAV (1/ft), ext. MC (percent), bed depth (ft)]
    # Note: Eventually we want to get rid of this and just use the dataframe.
    # This is legacy from the old parameter table json.
    sb40_dict = {}
    for key in params["key"]:
        row = params[params["key"] == key]
        sb40_dict[key] = [
            row["fbfm_code"].values[0],
            row["dry_load"].values[0],
            row["sav"].values[0],
            row["dead_fuel_extinction_moisture"].values[0],
            row["fuel_bed_depth_m"].values[0],
            row["duet_fuel_type"].values[0],
        ]

    return sb40_dict


def _get_sb40_arrays(sb40_keys: np.ndarray, sb40_dict: dict) -> np.ndarray:
    """
    Use a dictionary of bulk density and fuel types that correspond to SB40
    fuel models to assign those values across the study area.

    Fuel types are as follows:
    - 1: Predominantly grass. All cells with a GR, GS, or SH designation from SB40.
    - -1: Predominantly tree litter. All cells with a TL designation from SB40.
    - 0: Neither predominantly grass or tree litter. All other SB40 designations.

    Returns:
    3D np.ndarray:
    4 layers:
        1. fuel types
        2. bulk density values as calculated by fastfuels
        3. fuel moisture content values as calculated by fastfuels
        4. fuel height values as caluclated by fastfuels
    """
    val_idx = [5, 1, 3, 4]
    fuel_arr = np.zeros((4, sb40_keys.shape[0], sb40_keys.shape[1]))
    for i in range(len(val_idx)):
        layer_dict = {key: val[val_idx[i]] for key, val in sb40_dict.items()}
        layer = np.vectorize(layer_dict.get)(sb40_keys)
        fuel_arr[i, :, :] = layer

    return fuel_arr


def _reproject_polygon(poly: shapely.Polygon, input_epsg: int, target_epsg: int):
    coords = list(poly.exterior.coords)
    transformer = Transformer.from_crs(input_epsg, target_epsg, always_xy=True)
    transformed_coords = transformer.transform(*zip(*coords))
    transformed_poly = shapely.Polygon(list(zip(*transformed_coords)))

    return transformed_poly


def _delete_intermediate_files(directory: Path):
    # Name intermediate files
    temp = [
        "landfire_sb40.zip",
        "landfire_sb40.tif",
        "sb40_upsampled.tif",
        "sb40_cropped.tif",
    ]
    [Path(directory, file).unlink() for file in temp if Path(directory, file).exists()]
